-- Copyright (c) 2023-2024 Chris Pressey, Cat's Eye Technologies
-- This file is distributed under a BSD license.  See LICENSES directory:
-- SPDX-License-Identifier: LicenseRef-BSD-2-Clause-X-Fountain

module Language.Fountain.Grammar
  (
    Expr(Seq, Alt, Loop, Terminal, NonTerminal, Constraint),
    BacktrackMode(Deterministic, NonDeterministic),
    Production(Production, ntname, params, backtrackMode, constituents),
    Grammar(Grammar),
    depictExpr, depictExprs, depictProduction, depictGrammar, depictVars,
    depictChoice, depictChoices,
    startSymbol, productionExpr, productionParams, getFormals,
    getPreCondition, getLookahead
  ) where

import Data.List (intercalate)

import Language.Fountain.Value
import Language.Fountain.Constraint


type NTName = String

data BacktrackMode = Deterministic | NonDeterministic
    deriving (Show, Ord, Eq)

data Expr = Seq [Expr]
          | Alt (Maybe BacktrackMode) [Expr]
          --
          -- In the case of a Loop, there is a post-processing step
          -- that copies any constraints following the Loop, into the Loop
          -- itself.  This is to make the generator's job easier.
          --
          | Loop Expr [Constraint Value Expr]
          | Terminal Char
          | NonTerminal NTName [Variable]
          | Constraint (Constraint Value Expr)
    deriving (Show, Ord, Eq)

data Production = Production {
    ntname :: String,
    params :: [Variable],
    backtrackMode :: Maybe BacktrackMode,
    constituents :: Expr
} deriving (Show, Ord, Eq)

data Grammar = Grammar [Production]
    deriving (Show, Ord, Eq)

--
-- Pretty printing
--

depictExpr (Seq exprs) = "(" ++ (intercalate " " (map (depictExpr) exprs)) ++ ")"
depictExpr (Alt _ exprs) = "(" ++ (intercalate " | " (map (depictExpr) exprs)) ++ ")"
depictExpr (Loop expr _) = "{" ++ (depictExpr expr) ++ "}"
depictExpr (Terminal c) = "\"" ++ [c] ++ "\""
depictExpr (NonTerminal name vars) = name ++ depictVars vars
depictExpr (Constraint c) = "<. " ++ (depictConstraint c) ++ " .>"

depictExprs exprs = (intercalate ", " (map (depictExpr) exprs))

depictVars [] = ""
depictVars vars = "<" ++ (intercalate ", " (map (depictVar) vars)) ++ ">"

depictProduction p =
    (ntname p) ++ (depictVars $ params p) ++ (showBT $ backtrackMode p) ++ " ::= " ++ (depictExpr (constituents p)) ++ ";\n"
    where
        showBT Nothing = ""
        showBT (Just Deterministic) = "(!*)"
        showBT (Just NonDeterministic) = "(?*)"

depictGrammar (Grammar []) = ""
depictGrammar (Grammar (prod:rest)) = (depictProduction prod) ++ (depictGrammar $ Grammar rest)

depictChoice :: (Constraint Value Expr, Expr) -> String
depictChoice (cstr, expr) = (depictConstraint cstr) ++ ": " ++ (depictExpr expr)

depictChoices :: [(Constraint Value Expr, Expr)] -> String
depictChoices choices = (intercalate ", " (map (depictChoice) choices))

--
-- Accessors and utilities
--

startSymbol :: Grammar -> NTName
startSymbol (Grammar (prod : _)) = ntname prod
startSymbol (Grammar []) = error "No productions in grammar"

production :: NTName -> Grammar -> Production
production nt (Grammar (prod : rest)) =
    if nt == (ntname prod) then prod else production nt (Grammar rest)
production nt (Grammar []) = error ("Production '" ++ nt ++ "' not found")

productionExpr n g = constituents $ production n g
productionParams n g = params $ production n g

getFormals :: NTName -> Grammar -> [Variable]
getFormals nt (Grammar (prod : rest)) =
    if nt == (ntname prod) then params prod else getFormals nt (Grammar rest)
getFormals nt (Grammar []) = error ("Production '" ++ nt ++ "' not found")

--
-- Alt choices heed preconditions to prevent unnecessary backtracking.
--
getPreCondition :: Expr -> Constraint Value Expr
getPreCondition (Seq exprs) = getPreConditionPrefix Identity exprs
getPreCondition (Constraint c) = c
getPreCondition _ = Identity

getPreConditionPrefix c [] = c
getPreConditionPrefix c (Seq subseq:exprs) = getPreConditionPrefix c (subseq ++ exprs)
getPreConditionPrefix c (Constraint c':exprs) = getPreConditionPrefix (both c c') exprs
getPreConditionPrefix c _ = c

getLookahead :: Expr -> Maybe Char
getLookahead (Seq exprs) = getLookaheadPrefix exprs
getLookahead (Terminal ch) = Just ch
getLookahead _ = Nothing

getLookaheadPrefix [] = Nothing
getLookaheadPrefix (Seq subseq:exprs) = getLookaheadPrefix (subseq ++ exprs)
getLookaheadPrefix (Constraint _:exprs) = getLookaheadPrefix exprs
getLookaheadPrefix (Terminal ch:_) = Just ch
getLookaheadPrefix _ = Nothing
