-- Copyright (c) 2023-2024 Chris Pressey, Cat's Eye Technologies
-- This file is distributed under a BSD license.  See LICENSES directory:
-- SPDX-License-Identifier: LicenseRef-BSD-2-Clause-X-Fountain

module Language.Fountain.Parser (constructState, parseFrom, obtainResult) where

import Language.Fountain.Value
import Language.Fountain.Constraint
import Language.Fountain.Store (Store, empty, updateStore, filterStore, depictStore)
import Language.Fountain.Unify (applyConstraint)
import Language.Fountain.Grammar
import Language.Fountain.Loader (constructStore)


data ParseState = Parsing String (Store Value)
    deriving (Show, Ord, Eq)

type ParseResult = Either String ParseState

--
-- Utils
--

can (Right _) = True
can (Left _)  = False

expectTerminal :: Char -> ParseState -> ParseResult
expectTerminal tc (Parsing (c:cs) a)
    | c == tc   = Right $ Parsing cs a
    | otherwise = Left "no match"
expectTerminal _tc (Parsing [] _a) =
    Left "end of string"

applyParseConstraint :: Constraint Value Expr -> Grammar -> ParseState -> ParseResult
applyParseConstraint cstr g (Parsing s store) =
    case applyConstraint (handleLookahead) cstr store of
        Just store' ->
            Right $ Parsing s store'
        Nothing ->
            Left "constraint failed"
    where
        handleLookahead :: Constraint Value Expr -> Store Value -> Maybe (Store Value)
        handleLookahead (Lookahead e) sto =
            case parse g (Parsing s sto) e of
                Right (Parsing _ _) ->
                    Just sto
                Left _ ->
                    Nothing
        handleLookahead (IsGenerating) _ = Nothing
        handleLookahead other _ = error ("constraint " ++ (show other) ++ " should've already been handled") 

--
-- When getting applicable choices (for an Alt), first we narrow down
-- which ones are applicable according to their preconditions
-- (constraints at the front); then if the choice is still ambiguous,
-- we narrow down using token lookahead (which is parsing only --
-- generation does not need to do this.)
--

getApplicableChoicesByConstraint :: Grammar -> ParseState -> [Expr] -> [(Constraint Value Expr, Expr)]
getApplicableChoicesByConstraint g state alts =
    let
        preConditionedChoices = map (\x -> (getPreCondition x, x)) alts
        isApplicableChoice (c, _) = can $ applyParseConstraint c g state
    in
        filter (isApplicableChoice) preConditionedChoices

getApplicableChoices :: Grammar -> ParseState -> [Expr] -> [(Constraint Value Expr, Expr)]
getApplicableChoices g state alts =
    case getApplicableChoicesByConstraint g state alts of
        [] -> []
        [choice] -> [choice]
        choices@(_:_) ->
            --
            -- Constraints alone have not narrowed it down far enough, so
            -- start looking ahead at terminals.
            --
            let
                isApplicableChoice (_, cExpr) =
                    case (state, getLookahead cExpr) of
                        (Parsing (nextToken:_) _store, Just aToken) ->
                            aToken == nextToken
                        (_, _) ->
                            -- FIXME: maybe error message here
                            False
            in
                filter (isApplicableChoice) choices

--
-- Parser
--

parse :: Grammar -> ParseState -> Expr -> ParseResult

parse g state (Seq s) = parseSeq state s where
    parseSeq st [] = Right st
    parseSeq st (e : rest) =
        case parse g st e of
            Right st' -> parseSeq st' rest
            Left x    -> Left x

-- Hello, Mrs Backtracking Alternation!
parse g state (Alt (Just _) choices) = parseAlt state choices where
    -- Note, we try all the possibilities here, regardless of their preconditions.
    -- We also don't care if the backtracking is deterministic or nondeterministic.
    parseAlt _st [] = Left "no more choices"
    parseAlt st (e : rest) =
        case parse g st e of
            Left _    -> parseAlt st rest
            Right st' -> Right st'

-- Hello, Mrs Non-Backtracking Alternation!
parse g state (Alt Nothing choices) =
    parseAlt state $ getApplicableChoices g state choices
    where
        parseAlt _st [] = Left "no more choices"
        -- we ignore the constraint here because it will be found and applied when we descend into e
        parseAlt st [(_, e)] = parse g st e
        parseAlt (Parsing _str store) other =
            error (
                "Multiple pre-conditions are satisfied in Alt: " ++
                "[" ++ (depictChoices other) ++ "], with context: " ++ depictStore store
            )

parse g state (Loop l []) = parseLoop state l where
    parseLoop st e =
        case parse g st e of
            Right st' -> parseLoop st' e
            Left _    -> Right st

parse _g _state (Loop l (_:_)) = error ("Parsing can't handle decorated Loops: " ++ depictExpr l)

parse _g state (Terminal c) = expectTerminal c state

parse g (Parsing text store) (NonTerminal nt actuals) =
    let
        formals = getFormals nt g
        newStore = updateStore actuals formals store empty
        st' = Parsing text newStore
        expr' = productionExpr nt g
    in
        case parse g st' expr' of
            Right (Parsing text' modifiedStore) ->
                let
                    reconciledStore = updateStore formals actuals modifiedStore store
                in
                    Right $ Parsing text' reconciledStore
            Left x ->
                Left x

parse g state (Constraint cstr) = applyParseConstraint cstr g state

--
-- Usage interface
--

constructState :: Grammar -> String -> String -> [String] -> ParseState
constructState g start text givenParams =
    let
        givenStore = constructStore givenParams
        startParams = productionParams start g
        initialStore = filterStore givenStore startParams
    in
        Parsing text initialStore

parseFrom :: Grammar -> String -> ParseState -> ParseResult
parseFrom g start st = parse g st $ productionExpr start g

obtainResult :: ParseResult -> Either String String
obtainResult (Right (Parsing s _)) = Right s
obtainResult (Left other) = Left other
